# -*- coding: utf-8 -*-
""" `algorithm.py` contains the main implementation of the parallel (multiple-
chain) PoPEx algorithm. Mainly this concerns the two functions

    - :meth:`run_popex_mp`: Main implementation for PoPEx runs
    - :meth:`pred_popex_mp`: Main implementation for PoPEx predictions

that are used to sample models and predict results. A common file structure for
a PoPEx run together with some predictions is the following:

    | <path_res>$
    |  ├-- hd
    |     └-- hd_modXXXXXX.txt
    |  ├-- model
    |     └-- modXXXXXX.mod
    |  ├-- solution
    |      ├-- run_<name>_modXXXXXX
    |      └-- pred_<name>_modXXXXXX
    |  ├-- run_info.txt
    |  ├-- run_progress.txt
    |  ├-- pred_progress.txt
    |  ├-- popex.pop
    |  └-- problem.pb

The content of the different files is:

    - `hd_modXXXXXX.txt`:             Hard conditioning (without prior hd)
    - `modXXXXXX.mod`:                Pickled model object
    - `run_<name>_modXXXXXX.txt`:     Info from forward operator
    - `pred_<name>_modXXXXXX.txt`:    Info from prediction operator
    - `run_info.txt`:                 Info about the sampling
    - `run_progress.txt`:             Progress summary about the sampling
    - `pred_progress.txt`:            Progress summary about the predictions
    - `popex.pop`:                    Pickled PoPEx object
    - `problem.pb`:                   Pickled Problem object
"""

# -------------------------------------------------------------------------
#   Authors: Christoph Jaeggli, Julien Straubhaar and Philippe Renard
#   Year: 2019
#   Institut: University of Neuchâtel
#
#   Copyright (c) 2019 Christoph Jaeggli
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

# General imports
import warnings
import pickle
import time
import errno
import os
import numpy as np
import multiprocessing
from copy import deepcopy
from collections import deque
import bisect

# Package imports
import popex.isampl as isampl
import popex.popex_messages as msg
import popex.utils as utl
from popex.popex_objects import PoPEx, CatProb
from popex.cnsts import NP_MIN_TOL, RECMP_P_CAT_FREQ


def run_popex_mp(pb, path_res, path_q_cat,
                 ncmax=(20,), nmp=1, nmax=1000,
                 upd_hdmap_freq=1, upd_ls_freq=-1, si_freq=-1):
    """ `run_popex_mp` is the main implementation of the PoPEx algorithm.

    The algorithm expands a set of models until the defined stopping condition
    is fulfilled. A pre-requirement of the method is that we know a prior
    probability map that corresponds to a given set of categories (c.f.
    `q_cat`).


    Notes
    -----
    The prior and generation probability values should NEVER be considered as
    true probability values according to the true distribution. They should
    only be used in combination (as ratio r):

            `ratio(m) = rho(m) / phi(m)`.

    This ratio is important in the importance sampling framework where we want
    to compute weighted expectation values according to a set of generated
    models.


    Parameters
    ----------
    pb : Problem
        Defines the problem functions and parameters
    path_res : str
        Path to the 'results' folder
    path_q_cat : str
        Path to the prior probability maps `q_cat` such that the tuple of
        `nmtype` objects can be loaded under '<path_q_cat>q_cat.prob'
    ncmax : m-tuple
        Maximal number of conditioning points for each model type
    nmp : int
        Number of parallel processes to use
    nmax : int
        Number of maximal models (stopping condition)
    upd_hdmap_freq : int
        Defines the frequency for updating the HD maps (`kld` and `p_cat`)
    upd_ls_freq : int
         Defines the frequency for updating the learning scheme (-1 for no
         update)
    si_freq : int
         Defines the frequency of saving intermediate states (-1 for no
         intermediate saves)


    Returns
    -------
    None

    """

    # Initialization
    nmtype = pb.nmtype
    tst_run = time.time()
    popex = PoPEx(ncmax=ncmax, nmtype=nmtype, path_res=path_res)
    hd_prior_param_ind, hd_prior_param_val = pb.get_hd_pri()
    cond_mtype = [imtype for imtype in range(nmtype)
                  if ncmax[imtype] > 0]

    # Write run information to standard output
    print("\n  START 'RUN_POPEX_MP'\n")
    print('    ncmax   = {!r}'.format(ncmax))
    print('    nmax    = {:>6d}'.format(nmax))
    print('    nmp     = {:>6d}'.format(nmp))
    print("    hd_meth = '{}'\n".format(pb.meth_w_hd['name']))

    # Generate 'model' folder
    try:
        os.mkdir('{}model/'.format(path_res))
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
        pass

    # The map 'q_cat' must be a nmtype-tuple of 'CatProb' or 'None' objects
    print('    > load q_cat...', end='')
    with open('{}q_cat.prob'.format(path_q_cat), 'rb') as file:
        q_cat = pickle.load(file)
    for imtype in cond_mtype:
        if not isinstance(q_cat[imtype], CatProb):
            raise TypeError(imtype, type(q_cat[imtype]))
    print('done')

    # The kld map is zero initially (no conditioning data is selected)
    p_cat = deepcopy(q_cat)
    kld = utl.compute_kld(p_cat, q_cat)

    # Main PoPEx initialisations
    stop = False
    isim = 0
    sim_id_not_upd = []
    m_not_upd = []
    proc_mngr = deque([], maxlen=nmp)
    sim_id_mngr = []
    print('    > sample space...', end='')

    # Start PoPEx sampling by a pool of working processes
    with multiprocessing.Pool(processes=nmp, maxtasksperchild=100) as pool:
        while not stop:
            nmngr = len(proc_mngr)
            if nmngr < nmp and popex.nmod + nmngr < nmax:
                # Create and start a new working process
                args = (deepcopy(pb), deepcopy(popex), deepcopy(isim),
                        hd_prior_param_ind, hd_prior_param_val,
                        deepcopy(kld), deepcopy(p_cat), q_cat)
                proc_mngr.append(pool.apply_async(_run_process, args))
                isim += 1

            # Check if the process at the head of the queue has finished
            if proc_mngr[0].ready():
                # Remove process and get results
                proc = proc_mngr.popleft()
                (imod, model, ncmod, log_p_lik,
                 cmp_log_p_lik, log_p_pri, log_p_gen) = proc.get()

                # Extend control arrays for the update of 'p_cat'
                loc_upd = bisect.bisect(sim_id_not_upd, imod)
                sim_id_not_upd.insert(loc_upd, imod)
                m_not_upd.append(deepcopy(model))

                # Insert the model in the PoPEx structure
                loc_mngr = bisect.bisect(sim_id_mngr, imod)
                sim_id_mngr.insert(loc_mngr, imod)
                popex.insert_model(loc_mngr, imod, model, log_p_lik,
                                   cmp_log_p_lik, log_p_pri, log_p_gen, ncmod)

                # Write some information
                t_run = time.time() - tst_run
                t_mod = t_run / popex.nmod
                utl.write_run_info(pb, popex, imod, log_p_lik, cmp_log_p_lik,
                                   log_p_pri, log_p_gen, ncmod)
                _write_run_sum(pb, popex, nmax, t_run, t_mod)

                # Update 'kld' and 'p_cat' (if required)
                if len(cond_mtype) > 0:

                    if np.mod(popex.nmod, RECMP_P_CAT_FREQ) == 0:
                        # Frequently renew 'p_cat' entirely
                        w_hd = utl.compute_w_lik(popex=popex,
                                                 meth=pb.meth_w_hd)
                        p_cat = utl.compute_cat_prob(popex, w_hd)

                        # Compute the kld map with the new 'p_cat' map
                        kld = utl.compute_kld(p_cat, q_cat)

                        # 'p_cat' is "up to date" so we clean the arrays
                        sim_id_not_upd = []
                        m_not_upd = []

                    elif np.mod(popex.nmod, upd_hdmap_freq) == 0:
                        # Update 'p_cat' (which is much faster)
                        w_hd = utl.compute_w_lik(popex=popex,
                                                 meth=pb.meth_w_hd)
                        ind_not_upd = \
                            [i for i, id in enumerate(sim_id_mngr)
                             if id in sim_id_not_upd]
                        w_not_upd = w_hd[ind_not_upd]
                        sum_w_old = np.sum(w_hd) - np.sum(w_not_upd)
                        utl.update_cat_prob(p_cat, m_not_upd,
                                            w_not_upd, sum_w_old)

                        # Compute the kld map with the updated 'p_cat' map
                        kld = utl.compute_kld(p_cat, q_cat)

                        # 'p_cat' is "up to date" so we clean the arrays
                        sim_id_not_upd = []
                        m_not_upd = []

                # Update 'learning_scheme' (if required)
                if pb.learning_scheme and cmp_log_p_lik:
                    n_cmp = np.sum(popex.cmp_log_p_lik)
                    if all((upd_ls_freq > 0, np.mod(n_cmp, upd_ls_freq)==0)):
                        pb.learning_scheme.train(popex)

                # Update stopping condition
                if popex.nmod >= nmax:
                    stop = True

                # Save intermediate PoPEx result
                if all([si_freq > 0,
                        np.mod(popex.nmod, si_freq) == 0,
                        stop is False]):
                    with open('{}popex.pop'.format(popex.path_res), 'wb')\
                            as file:
                        pickle.dump(popex, file)

            # Otherwise send it to the back of the queue
            else:
                proc_mngr.rotate(-1)
    print('done')

    # Save final PoPEx result
    print('    > save PoPEx set...', end='')
    with open('{}popex.pop'.format(popex.path_res), 'wb') as file:
        pickle.dump(popex, file)
    print('done')

    # Save Problem
    print('    > save problem...', end='')
    with open('{}problem.pb'.format(popex.path_res), 'wb') as file:
        pickle.dump(pb, file)
    print('done')

    print("\n  END 'RUN_POPEX_MP'\n")


def _run_process(pb, popex, imod,
                 hd_prior_param_ind, hd_prior_param_val,
                 kld, p_cat, q_cat):
    """ `_run_process` is the method that can be used to sample a new model in
    the PoPEx procedure. It mainly generates a new model and computes the
    corresponding log-likelihood value.


    Parameters
    ----------
    pb : Problem
        Problem definition
    popex : PoPEx
        PoPEx main structure
    imod : int
        Model index
    hd_prior_param_ind : m-tuple
        Prior hard conditioning indices (cf. :meth:`pb.get_hd_pri`)
    hd_prior_param_val : m-tuple
        Prior hard conditioning values (cf. :meth:`pb.get_hd_pri`)
    kld : m-tuple
        Kullback-Leibler divergence map (cf. :meth:`pb.compute_kld`)
    p_cat : m-tuple
        Weighted category probabilities
    q_cat : m-tuple
        Prior category probabilities


    Returns
    -------
    imod : int
        Model index
    model : m-tuple
        Tuple of ``Mtype`` instances that define the new model
    ncmod : m-tuple
        Number of hard conditioning for each model type
    log_p_lik : float
        Log-likelihood value
    cmp_log_p_lik : bool
        Log-likelihood was computed (True) or predicted (False)
    log_p_pri : float
        Log-prior probability of the model
    log_p_gen : float
        Log-sampling probability of the model

    """

    np.random.seed(pb.seed + imod)
    meth_w_hd = pb.meth_w_hd

    # Number of conditioning (is zero everywhere if sum(w_hd) = 0)
    ncmod_tmp = utl.compute_ncmod(popex, meth_w_hd)

    # Hard conditioning
    hd_i_param_ind, hd_i_param_val, hd_prior, hd_generation = \
        utl.generate_hd(popex, meth_w_hd, ncmod_tmp, kld, p_cat, q_cat)
    hd_param_ind, hd_param_val = \
        utl.merge_hd(hd_prior_param_ind, hd_i_param_ind,
                     hd_prior_param_val, hd_i_param_val)

    # The number of hard conditioning data may has changed (duplicated loc.)
    ncmod = tuple([len(hd_ind) if hd_ind is not None else 0
                   for hd_ind in hd_i_param_ind])

    # Generate model
    model = pb.generate_m(hd_param_ind, hd_param_val, imod)

    # Write hd information (!!! without prior hd !!!)
    utl.write_hd_info(popex, imod, hd_i_param_ind, hd_i_param_val)

    # Compute log-likelihood (if learning_scheme is None or with p_eval)
    if not pb.learning_scheme or \
            np.random.rand(1) < pb.learning_scheme.compute_p_eval_for(model):
        log_p_lik = pb.compute_log_p_lik(model, imod)
        cmp_log_p_lik = True
    # Predict likelihood (otherwise)
    else:
        log_p_lik = pb.learning_scheme.learn_value_of(model)
        cmp_log_p_lik = False

    # Test positivity of likelihood value
    if np.exp(log_p_lik) < -NP_MIN_TOL:
        raise ValueError(msg.err4001(list(imod)))

    # Compute sampling ratio
    log_p_pri = pb.compute_log_p_pri(model, hd_prior, hd_i_param_ind)
    log_p_gen = pb.compute_log_p_gen(model, hd_generation, hd_param_ind)

    return imod, model, ncmod, log_p_lik, cmp_log_p_lik, log_p_pri, log_p_gen


def _write_run_sum(pb, popex, nmax, t_popex, t_mod):
    """ `_write_run_sum` writes a document that summarized the overall
    progress of the popex run at 'popex.path_res'.


    The file structure is as followes:
        | <popex.path_res>$
        |    └- run_progress.txt


    Parameters
    ----------
    pb : Problem
        Problem definition
    popex : PoPEx
        PoPEx main structure
    nmax : int
        Number of maximal models
    t_popex : float
        Total time elapsed
    t_mod : float
        Average time per model


    Returns
    -------
    None

    """

    # Generic constants
    bar_width = 40

    # Path of the result location
    path_res = popex.path_res

    # Compute n_e diagnostics
    with warnings.catch_warnings(record=True) as _:
        ne_l    = isampl.ne(utl.compute_w_lik(popex=popex,
                                              meth={'name': 'exact'}))
        ne_w_hd = isampl.ne(utl.compute_w_lik(popex=popex,
                                              meth=pb.meth_w_hd))
        ne_w_l  = isampl.ne(utl.compute_w_pred(popex=popex,
                                               nw_min=0,
                                               meth={'name': 'exact'}))

    # Compute time information (t_popex, t_mod, t_est)
    progress = popex.nmod / nmax
    m_popex, s_popex = divmod(t_popex, 60)
    h_popex, m_popex = divmod(m_popex, 60)
    m_mod, s_mod = divmod(t_mod, 60)
    h_mod, m_mod = divmod(m_mod, 60)
    t_est = t_popex / progress
    m_est, s_est = divmod(t_est, 60)
    h_est, m_est = divmod(m_est, 60)

    # Write into a file 'iteration_info.txt'
    with open('{}/run_progress.txt'.format(path_res), 'w+') as file:
        # Start with empty line
        file.write('\n')

        # Write title
        file.write('RUN PROGRESS SUMMARY\n')
        file.write('--------------------\n')
        file.write('\n')

        # Print status bar
        nfull = int(bar_width * progress)
        nempty = bar_width - nfull
        file.write('  Status\n')
        file.write('    n_mod       = {:7d} / {:7d}\n'.format(popex.nmod, nmax))
        file.write('    [{:s}{:s}]'.format('-' * nfull, ' ' * nempty))
        file.write(' {:3.0f}%\n'.format(progress * 100))
        file.write('\n')

        # Write n_e
        file.write('  Diagnostics\n')
        file.write('    ne(L)       = {:9.1f}\n'.format(ne_l))
        file.write('    ne(w_hd)    = {:9.1f}\n'.format(ne_w_hd))
        file.write('    ne(w_pred)  = {:9.1f}\n'.format(ne_w_l))
        file.write('\n')

        # Write compute/predict stats
        n_comp = np.sum(popex.cmp_log_p_lik)
        n_pred = popex.nmod - n_comp
        file.write('  Computed or Predicted\n')
        file.write('    n_cmp       = {:7d}\n'.format(n_comp))
        file.write('    n_prd       = {:7d}\n'.format(n_pred))
        file.write('\n')

        # Write times
        file.write('  Time\n')
        file.write('    T_el        = {:4.0f}[h] {:2.0f}[m] {:5.2f}[s]\n'
                   .format(h_popex, m_popex, s_popex))
        file.write('    T / mod     = {:4.0f}[h] {:2.0f}[m] {:5.2f}[s]\n'
                   .format(h_mod, m_mod, s_mod))
        file.write('    T_tot       = {:4.0f}[h] {:2.0f}[m] {:5.2f}[s]\n'
                   .format(h_est, m_est, s_est))

        # End with empty line
        file.write('\n')


def pred_popex_mp(pred, path_res, nmp=1):
    """ `pred_popex_mp` is the main implementation for computing predictions
    from a PoPEx sampling.


    Notes
    -----
    The computations are supposed to be independent, such that they can be
    computed in parallel.


    Parameters
    ----------
    pred : Prediction
        Defines the prediction functions and parameters
    path_res : str
        Path for loading the PoPEx results
    nmp : int
        Number of parallel processes


    Returns
    -------
    None

    """
    # Start PoPEx predictions
    tst_pred = time.time()

    # Load PoPEx results
    with open(path_res + 'popex.pop', 'rb') as file:
        popex = pickle.load(file)  # type: PoPEx

    # Compute prediction indices
    w_pred = utl.compute_w_pred(popex=popex,
                                nw_min=pred.nw_min,
                                meth=pred.meth_w_pred)
    ind_pred = utl.compute_subset_ind(pred.wfrac_pred, w_pred)
    npred = len(ind_pred)

    # Write some information
    print("\n  START 'PRED_POPEX_MP'\n")
    print('    nmp        = {:>6}'.format(nmp))
    print('    npred      = {:>6}'.format(npred))
    print("    pred_meth  = '{}'\n".format(pred.meth_w_pred['name']))

    # Main prediction loop
    stop = False
    ipred = 0
    ndone = 0
    proc_mngr = deque([], maxlen=nmp)
    print('    > predict data...', end='')
    with multiprocessing.Pool(processes=nmp, maxtasksperchild=100) as pool:
        while not stop:
            if len(proc_mngr) < nmp and ipred < npred:
                # Create new process
                imod = ind_pred[ipred]
                args = (deepcopy(pred), deepcopy(popex), imod)
                proc_mngr.append(pool.apply_async(_pred_process, args))
                ipred += 1

            # Check if the process at the head of the queue has finished
            if proc_mngr[0].ready():
                # Remove process
                proc = proc_mngr.popleft()
                ipred_proc = proc.get()
                ndone += 1

                # Write some information
                t_pred = time.time() - tst_pred
                _write_pred_sum(path_res, ipred_proc, ndone, npred, t_pred)

                # Update stopping condition
                if len(proc_mngr) == 0 and ipred >= npred:
                    stop = True
            else:
                # Send process to the back of the queue
                proc_mngr.rotate(-1)
    print('done')

    # Save Prediction
    print('    > save prediction...', end='')
    with open('{}pred.pred'.format(popex.path_res), 'wb') as file:
        pickle.dump(pred, file)
    print('done')

    print("\n  END 'PRED_POPEX_MP'\n")


def _pred_process(pred, popex, imod):
    """ `_pred_process` is the method that can be used to compute a prediction
    from one model in a PoPEx sampling.


    Parameters
    ----------
    pred : Prediction
        Defines the prediction functions and parameters
    popex : PoPEx
        PoPEx main structure
    imod : int
        Model index


    Returns
    -------
    None
    """
    pred.compute_pred(popex, imod)
    return imod


def _write_pred_sum(path_res, ipred, ndone, ntot, t_pred):
    """ `_write_pred_sum` writes a text file that summarized the overall
    progress of the predictions at 'popex.path_res'.


    The file structure is as followes:
        | <popex.path_res>$
        |    └- pred_progress.txt


    Parameters
    ----------
    path_res : str
        Path to where the results should be saved
    ipred : int
        Index of last predictions
    ndone : int
        Number of computed predictions
    ntot : int
        Number of total predictions
    t_pred : float
        Total time elapsed


    Returns
    -------
    None
    """
    # Generic constants
    bar_width = 40

    # Compute time information (t_pred, t_mod, t_est)
    progress = ndone / ntot
    t_mod = t_pred / ndone
    m_pred, s_pred = divmod(t_pred, 60)
    h_pred, m_pred = divmod(m_pred, 60)
    m_mod, s_mod = divmod(t_mod, 60)
    h_mod, m_mod = divmod(m_mod, 60)
    t_est = t_pred / progress
    m_est, s_est = divmod(t_est, 60)
    h_est, m_est = divmod(m_est, 60)

    # Write into a file 'iteration_info.txt'
    with open('{}pred_progress.txt'.format(path_res), 'w+') as file:
        # Write title
        file.write('PREDICTION PROGRESS SUMMARY\n')
        file.write('---------------------------\n')
        file.write('\n')

        # Print status bar
        nfull = int(bar_width * progress)
        nempty = bar_width - nfull
        file.write('  Status\n')
        file.write('    n_mod    = {:6d} / {:6d}    (last: {:6d})\n'
                   .format(ndone, ntot, ipred))
        file.write('    [{:s}{:s}]'.format('-' * nfull, ' ' * nempty))
        file.write(' {:3.0f}%\n'.format(progress * 100))
        file.write('\n')

        # Write times
        file.write('  Time\n')
        file.write('    T_el     = {:4.0f}[h] {:2.0f}[m] {:5.2f}[s]\n'
                   .format(h_pred, m_pred, s_pred))
        file.write('    T / mod  = {:4.0f}[h] {:2.0f}[m] {:5.2f}[s]\n'
                   .format(h_mod, m_mod, s_mod))
        file.write('    T_tot    = {:4.0f}[h] {:2.0f}[m] {:5.2f}[s]\n'
                   .format(h_est, m_est, s_est))



""" 'popex_messages.py' contains a collection of messages that are used for
warnings and errors appearing in the popex package.
"""

# ------------------------------- 1XXX utils.py -------------------------------
# Warnings
warn1001 = "\nWARNING 1001: The sum of the weights was equal to 0. The " \
           "weights have been redefined to be uniformly equal to 1/n.\n"


def warn1002(imtype):
    return "\nWARNING 1002: For the model type {} a location x has been " \
           "detected where q_i(x) = 0 and p_i(x) > 0.\nIf q was computed " \
           "from a set of N models, try to increase N.\n".format(imtype)


warn1003 = "\nWARNING 1003: The sum of the kld was equal to 0. The " \
           "values have been redefined to be uniformly equal to 1/n.\n"


warn1004 = "\nWARNING 1004: The sum of the likelihood was equal to 0. The " \
           "values have been redefined to be uniformly equal to 1/n.\n"


# Errors
err1001 = "\nERROR 1001: Negative weight(s) detected."


def err1002(imtype, ind):
    return "\nERROR 1002: Negative Kullback-Leibler-Divergence detected for " \
           "model type {} at location index {}.\n".format(imtype, ind)


def err1003(imod):
    return "\nERROR 1003: HD-Weight value(s) {} is/are negative.".format(imod)


def err1006(first, last):
    return "\nERROR 1006: The set of models [{}, {}) based on which the" \
           "category probability should be computed is empty."\
           .format(first, last)


def err1007(first, last):
    return '\nERROR 1007: The weights between {} and {} have not properly ' \
           'been normalized'.format(first, last)


err1008 = '\nERROR 1008: Probability value at hard conditioning location is 0'


err1009 = '\nERROR 1009: Wrong length of HD parameter indices.'


def err1010(ind):
    return "\nERROR 1010: The log-values {!r} are positive.".format(ind)


err1011 = "\nERROR 1011: The input parameter 'm_new' in " \
          "'utils.update_cat_prob' must be of type 'list'\n"

err1012 = "\nERROR 1012: The input parameter 'w_new' in " \
          "'utils.update_cat_prob' must be of type 'np.ndarray'\n"

err1013 = "\nERROR 1013: The input parameters 'm_new' and 'w_new' in " \
          "'utils.update_cat_prob' must have same length\n"

# ---------------------------- 2XXX popex_objects -----------------------------
# Warnings


# Errors
def err2001(ind, vals, cat):
    return '\nERROR 2001: param_val{} = {} was/were not properly ' \
           'categorized by the categories in {!r}.'.format(ind, vals, cat)


err2002 = "\nERROR 2003: Attribute 'param_val' must be a 1 dimensional array" \
          " or 'None'."


err2003 = "\nERROR 2004: Attribute 'param_val' must be a 2 dimensional array" \
          " or 'None'."


def err2004(ncat):
    return "\nERROR 2004: Too many categories ({})! Redefine default numpy " \
           "type of 'param_cat' in source code of " \
           "'popex.popex_objects.CatParam'"


err2005 = "\nERROR 2005: Attribute 'categories' must be a list."

# ------------------------------- 3XXX problem --------------------------------
# Warnings


# Errors
err3001 = "\nERROR 3001: Average value of the KLD map is larger then KLD_MAX."


# ------------------------------ 4XXX algorithm -------------------------------
# Warnings


# Errors
def err4001(imod):
    return "\nERROR 4001: Likelihood value(s) {!r} is/are negative."\
           .format(imod)


def err4002(imod, val):
    return "\nERROR 4002: Log-prior probability value {} is equal to {}."\
           .format(imod, val)


def err4003(imod, val):
    return "\nERROR 4003: Log-generation probability value {} is equal to {}."\
           .format(imod, val)


def err4004(imod, n_is, n_should):
    return '\nERROR 4004: Model {:d} defines {:d} instead of {:d} model types'\
           .format(imod, n_is, n_should)


def err4005(imod, imtype, type_i):
    return "\nERROR 4005: Wrong instance in model {:d}. All models types " \
           "must subclass 'MType', but type(m[{:d}] = {!r}"\
           .format(imod, imtype, type_i)


def err4006(imtype, type_i):
    return "\nERROR 4006: Model instances of type {!r} can not be " \
           "conditioned. Put ncmax[{:d}] equal to 0.".format(type_i, imtype)


def err4007(imod, hd_meth):
    return "\nERROR 4007: Log-likelihood value of model {} is positive, what" \
           " is not possible for 'hd_meth' = {}."\
           .format(imod, hd_meth)


def err4008(imtype, map_type):
    return "\nERROR 4008: The prior probability map in 'q_cat[{}]' is of " \
           "type {!r}.".format(imtype, map_type)


# ----------------------------- 5XXX predictions ------------------------------
# Warnings


# Errors
def err5001(sflag):
    return "\nERROR 4001: Not implemented save option for flag = {}"\
           .format(sflag)



# -*- coding: utf-8 -*-
""" `isample.neff` is a module that implements the most common importance
sampling diagnostics used in the PoPEx procedure. Namely these are

    - :meth:`ne`            Effective number of weights for estimating the
      expectation
    - :meth:`ne_var`        Effective number of weights for estimating the
      variance
    - :meth:`ne_gamma`      Effective number of weights for estimating the
      skewness
    - :meth:`alpha`         Optimization for finding the weight correction
      exponent
    - :meth:`correct_w`     Computes the set of corrected weights
"""

# -------------------------------------------------------------------------
#   Authors: Christoph Jaeggli, Julien Straubhaar and Philippe Renard
#   Year: 2018
#   Institut: University of Neuchatel
#
#   Copyright (c) 2018 Christoph Jaeggli
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

import warnings
import numpy as np
from scipy import optimize
import popex.popex_messages as msg
from popex.cnsts import NP_MIN_TOL, ALPHA_MIN, ALPHA_MAX


def ne(weights):
    """ `ne` computes the effective number of weights.

    Kish's effective number of weights is computed as

        `n_e(w_1, ..., w_n) = ( \sum w_i )^2 / ( \sum w_i^2 )`.


    Parameters
    ----------
    weights : ndarray, shape=(n,)
        Set of weights.

    Returns
    -------
    float
        Effective number of weights.

    """
    # Raise numpy over-/underflow errors
    old_settings = np.seterr(over='raise', under='raise')

    # Treat extreme values
    w = np.array(weights)
    try:
        w_sq = w ** 2
    except FloatingPointError as err:
        first_word = str(err).split()[0]
        if first_word == 'overflow':
            w[w >= NP_MIN_TOL] /= np.max(w)
            w[w < NP_MIN_TOL] = 0
        elif first_word == 'underflow':
            w[w < NP_MIN_TOL] = 0
        else:
            raise FloatingPointError(err)
        w_sq = w ** 2

    # Treat zero division
    if np.sum(w_sq) < NP_MIN_TOL:
        w = np.ones_like(w)
        w_sq = w
        warnings.warn(msg.warn1001)

    # Set numpy settings back
    np.seterr(**old_settings)

    # Return effective number of weights
    return (np.sum(w) ** 2) / np.sum(w_sq)


def ne_var(weights):
    """ `ne_var` computes the effective number of weights for estimating the
    variance.

    The effective number of weights for an empirical estimation of the variance
    is computed as:

        `n_e(w_1, ..., w_n) = ( \sum w_i^2 )^2 / ( \sum w_i^4 )`.


    Parameters
    ----------
    weights : ndarray, shape=(n,)
        Set of weights.

    Returns
    -------
    float
        Effective number of weights (for the variance).

    """
    # Raise numpy overflow errors
    old_settings = np.seterr(over='raise', under='raise')

    w = np.array(weights)
    if np.sum(w ** 2) < NP_MIN_TOL:
        w = np.ones_like(w)
        warnings.warn(msg.warn1001)

    # Set numpy settings back
    np.seterr(**old_settings)

    # Return effective number of weights
    return (np.sum(w ** 2) ** 2) / np.sum(w ** 4)


def ne_gamma(weights):
    """ `ne_gamma` computes the effective number of weights for the skewness.

    The effective number of weights for estimating the skewness is computed as:

        `n_e(w_1, ..., w_n) = ( \sum w_i^2 )^3 / ( ( \sum w_i^3 )^2 )`.


    Parameters
    ----------
    weights : ndarray, shape=(n,)
        Set of weights.

    Returns
    -------
    float
        Effective number of weights (for the skewness).

    """
    # Raise numpy overflow errors
    old_settings = np.seterr(over='raise', under='raise')

    w = np.array(weights)
    if np.sum(w ** 2) < NP_MIN_TOL:
        w = np.ones_like(w)
        warnings.warn(msg.warn1001)

    # Set numpy settings back
    np.seterr(**old_settings)

    # Return effective number of weights
    return (np.sum(w ** 2) ** 3) / (np.sum(w ** 3) ** 2)


def alpha(weights, theta, a_init=1):
    """ `alpha` computes the best `alpha` for lowering the skewness of the
    weights.


    Let `k_1` be the number of zero weights and `k_2` is the number of weights
    that attain the maximum value in 'weights'. For a given theta in the
    interval `(k_2, n-k_1)`, this functions finds the unique `alpha` such that

        `neff.ne(weights ** alpha) = theta`.

    The alpha is found by using the 'L-BFGS-B' optimization algorithm of
    `scipy.optimize` with initial value equal to `a_init`. We set upper and
    lower bounds to `ALPHA_MIN` and `ALPHA_MAX`, respectively. The maximum
    number of iterations is 15.

    Parameters
    ----------
    weights : ndarray, shape=(n,)
        Set of weights
    theta : float
        A positive value for the effective number of weights
    a_init : float
        Initial guess for the optimization problem


    Returns
    -------
    float
        `alpha` value

    """
    # Set numpy error behaviour
    old_settings = np.seterr(over='raise', under='raise')
    w = np.array(weights)
    # w[w < NP_MIN_TOL] = 0

    # Define optimization function
    def opt_fun(a):
        try:
            w_a = w ** a
        except FloatingPointError:
            if a < 1:
                return (np.sum(w > 0) - theta) ** 2
            else:
                return (np.sum(w == np.max(w)) - theta) ** 2
        return (ne(w_a) - theta) ** 2

    # Minimize optimization function
    nnz = np.sum(w > 0)
    nmax = np.sum(w == np.max(w))
    if theta > nmax:
        if theta < nnz:
            opt_obj = optimize.minimize(opt_fun, np.array([a_init]),
                                        method='L-BFGS-B',
                                        bounds=[(ALPHA_MIN, ALPHA_MAX)],
                                        options={'disp': False,
                                                 'maxiter': 10})

            a_ret = opt_obj['x'][0]
        else:
            a_ret = ALPHA_MIN
    else:
        a_ret = ALPHA_MAX
    # Set back to old settings
    np.seterr(**old_settings)
    return a_ret


def correct_w(weights, ne_w_corr):
    """ `correct_w` is a function that lowers the skewness of the weights.

    In addition to the method :meth:`alpha` it computes whether the correction
    is possible and treats the exceptions accordingly.


    Parameters
    ----------
    weights : ndarray, shape=(n,)
        Set of weights
    ne_w_corr : float
        A positive value for the effective number of weights


    Returns
    -------
    w_corr : ndarray, shape=(n,)
        Corrected set of weights that is also normalized

    """
    a_val = alpha(weights, ne_w_corr, a_init=1)

    if np.abs(a_val - ALPHA_MIN) < NP_MIN_TOL:
        # Set all positive value to 1
        w_corr = np.zeros_like(weights)
        w_corr[weights >= NP_MIN_TOL] = 1

    elif np.abs(a_val - ALPHA_MAX) < NP_MIN_TOL:
        # Set all values that attain the max to 1
        w_corr = np.zeros_like(weights)
        w_corr[weights == np.max(weights)] = 1

    else:
        # Compute the power weights
        w_corr = weights ** a_val

    # Check for division error
    if np.sum(w_corr) < NP_MIN_TOL:
        w_corr = np.ones_like(w_corr)
    w_corr /= np.sum(w_corr)

    return w_corr



# -*- coding: utf-8 -*-
""" `utils.py` contains utilities for performing a PoPEx sampling and computing
predictions:

Category probabilities and kld maps
    - :meth:`compute_cat_prob`:   Computes probability maps according to
      categories
    - :meth:`update_cat_prob`:    Updates probability maps according to
      categories
    - :meth:`compute_entropy`:    Computes entropy of a probability map
    - :meth:`compute_kld`:        Computes kld of two probability maps

Hard conditioning data
    - :meth:`generate_hd`:        Computes the new hard conditioning data
    - :meth:`merge_hd`:           Merges prior and new hard conditioning data
    - :meth:`compute_ncmod`:      Computes the number of conditioning points per
      model type
    - :meth:`compute_w_lik`:      Computes the likelihood weights (used for the
      hard conditioning maps)

Generic functions
    - :meth:`compute_w_pred`:     Computes the weights for the predictions
    - :meth:`compute_subset_ind`: Computes the smallest number of indices that
      cover a given percentage of a total weight
    - :meth:`write_hd_info`:      Writes/saves hd information about each model
    - :meth:`write_run_info`:     Appends information about models to run info
      file
"""

# -------------------------------------------------------------------------
#   Authors: Christoph Jaeggli, Julien Straubhaar and Philippe Renard
#   Year: 2019
#   Institut: University of Neuchâtel
#
#   Copyright (c) 2019 Christoph Jaeggli
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

# General imports
import os
import errno
import warnings
import numpy as np
import pickle
from copy import deepcopy

# Package imports
import popex.isampl as isampl
import popex.popex_messages as msg
from popex.popex_objects import PoPEx, CatMType, ContParam, CatProb
from popex.cnsts import NP_MIN_TOL


# Category probabilities and kld maps
def compute_cat_prob(popex, weights, start=-1, stop=-1):
    """ `compute_cat_prob` computes the weighted category probabilities.

    The models are obtained from `popex.model` and weighted by `weights`.

    Parameters
    ----------
    popex : PoPEx
        PoPEx main structure (cf `popex_objects.PoPEx`)
    weights : ndarray, shape=(nmod,)
        Relative weights of the models
    start : int
        Defines the first model to take into account:

        - -1: For starting at 0
        - N: For starting at `max(N, 0)`
    stop : int
        Defines the last model to take into account

        - -1: For stopping at `popex.nmod`
        - N: For stopping at `min(N, popex.nmo)`

    Returns
    -------
    m-tuple
        A tuple of instances that describes the category probabilities for all
        categorical model types.

        If a model type i is not a subclass of ``CatMType``, the corresponding
        map is set to ``None``. If a model is given by (`CatModel_1`, ...,
        `CatModel_m`) and the model values in `CatModel_i` are subdivided into
        `ncat_i` categories, then the return value is a (`CatProb_1`, ...,
        `CatProb_m`) tuple where `return[i].param_val` is a ``ndarray`` with
        `shape=(nparam_i, ncat_i)`.

    """
    nmtype = popex.nmtype

    # Evaluate starting and ending point
    if start == -1:
        first = 0
    else:
        first = max(start, 0)
    if stop == -1:
        last = popex.nmod
    else:
        last = min(stop, popex.nmod)
    if first >= last:
        raise ValueError(msg.err1006(first, last))

    # Normalize weights
    if np.any(weights[first:last] < -NP_MIN_TOL):
        raise ValueError(msg.err1001)
    if np.sum(weights[first:last]) >= NP_MIN_TOL:
        w_nrm = weights[first:last] / np.sum(weights[first:last])
    else:
        w_nrm = np.ones(last - first) / (last - first)
        warnings.warn(msg.warn1001)
    if np.abs(np.sum(w_nrm[first:last]) - 1) > NP_MIN_TOL:
        raise ValueError(msg.err1007(first, last))

    # Instantiate p_cat
    mod = deepcopy(popex.model[first])
    if isinstance(mod, str):
        with open(popex.path_res + mod, 'rb') as mfile:
            mod = pickle.load(mfile)
    p_cat = [None for _ in range(nmtype)]
    cat_mtype = [imtype for imtype in range(nmtype)
                 if isinstance(mod[imtype], CatMType)]
    for imtype in cat_mtype:
        p_cat[imtype] = CatProb(param_val=np.zeros((mod[imtype].nparam,
                                                    mod[imtype].ncat)),
                                categories=mod[imtype].categories)

    # Compute category probabilities
    for imod in range(first, last):
        if imod > first:
            mod = deepcopy(popex.model[imod])
            if isinstance(mod, str):
                with open(popex.path_res + mod, 'rb') as mfile:
                    mod = pickle.load(mfile)
        for imtype in cat_mtype:
            for icat in range(mod[imtype].ncat):
                p_cat[imtype].param_val[:, icat] += \
                    (mod[imtype].param_cat == icat) * w_nrm[imod]
    return tuple(p_cat)


def update_cat_prob(p_cat, m_new, w_new, sum_w_old):
    """ `update_cat_prob` updates (in place) the category probabilities.

    If `p_cat_old` represents the old category probability maps, then we have

        `p_cat_new = [sum_w_old*p_cat_old + \sum_i w_new_i*1(m_new_i)]
        / [sum_w_old + sum(w_new)]`

    where `1(m_new_i)` is the categorical indicator of the model `i`.


    Parameters
    ----------
    p_cat : m-tuple
        Tuple of categorical probability maps (cf. :meth:`compute_cat_prob`)
    m_new : list
        List of m-tuples defining a set of `nmod` models
    w_new : ndarray, shape=(nmod,)
        Weights associated to the new models
    sum_w_old : float
        Old weight normalization constant


    Returns
    -------
    None

    """

    # Check input types
    if not isinstance(m_new, list):
        raise ValueError(msg.err1011)
    if not isinstance(w_new, np.ndarray):
        raise ValueError(msg.err1012)
    elif not len(m_new) == w_new.size:
        raise ValueError(msg.err1013)

    # Check positivity of weights
    if any(w < 0 for w in w_new):
        raise ValueError(msg.err1001)

    # p_cat stays unchanged if the new weight is zero
    sum_w_new = np.sum(w_new)
    nmtype = len(p_cat)
    if sum_w_new >= NP_MIN_TOL:
        # Generate new p_cat instance (based on new models)
        popex_new = PoPEx(model=m_new, nmtype=nmtype)
        p_cat_new = compute_cat_prob(popex_new, w_new)

        # Update weights
        w_p_cat = (sum_w_old / (sum_w_old+sum_w_new))
        w_p_cat_new = (sum_w_new / (sum_w_old+sum_w_new))

        # p_cat[imtype] may be 'None' (for unconditioned model types)
        cat_mtype = [imtype for imtype in range(nmtype)
                     if not p_cat[imtype] is None]

        # Update p_cat
        for imtype in cat_mtype:
            p_cat[imtype].param_val = p_cat[imtype].param_val*w_p_cat \
                                      + p_cat_new[imtype].param_val*w_p_cat_new
    return None


def compute_entropy(p_cat):
    """ `compute_entropy` computes the entropy of `p_cat`.

    The entropy of a discrete probability distributions `p = (p_1, ...,p_s)` is

        `H(p) = -\sum_{i=1}^s p_i log( p_i )`.

    Therefore, if the probability map `p_cat` is a ``m-tuple`` such that
    `p_cat[i].param_val` is an ``ndarray`` of `shape=(nparam_i, nfac_i)`, the
    entropy is also an ``m-tuple`` where `H[i].param_val` being an ``ndarray``
    of `shape=(nparam,)`.

    Notes
    -----
    Note that `t*log(t) -> 0` as `t -> 0`. Therefore, `H(x) = 0` wherever
    `p_i(x) = 0`.


    Parameters
    ----------
    p_cat : m-tuple
        Tuple of ``CatProb`` instances with `p_cat[i].param_val` being an
        ``ndarray`` of `shape=(nparam_i, nfac_i)`


    Returns
    -------
    m-tuple
        Tuple of entropy maps

        `return[i]` : ``None`` or ``ContParam``
            Return value `i` is ``None`` if `p_cat[i]` is ``None``, otherwise
            it is an instance of ``ContParam``

    """
    nmtype = len(p_cat)

    # Compute H only not-None probability maps
    H = [None for _ in range(nmtype)]
    cat_mtype = [imtype for imtype in range(nmtype)
                 if not p_cat[imtype] is None]

    # Iteratively compte H for each mtype
    for imtype in cat_mtype:
        H[imtype] = ContParam()
        p_i = p_cat[imtype].param_val

        # Compute KLD
        nparam, ncat = p_cat[imtype].param_val.shape
        H[imtype].param_val = np.zeros((nparam,))
        for icat in range(ncat):
            nnz_ind = p_i[:, icat] > NP_MIN_TOL
            H[imtype].param_val[nnz_ind]\
                -= p_i[nnz_ind, icat] * np.log(p_i[nnz_ind, icat])

    return tuple(H)


def compute_kld(p_cat, q_cat):
    """ `compute_kld` computes the Kullback-Leibler divergence (KLD) between
    two category probability maps `p_cat` and `q_cat`.

    The KLD between two discrete probability distributions `p = (p_1, ...,p_s)`
    and `q = (q_1, ...,q_s)` is

        `KLD(p||q) = \sum_{i=1}^s p_i log( p_i / q_i)`.

    Therefore, if the probability maps `p_cat` and `q_cat` are ``m-tuples`` such
    that `p_cat[i].param_val` and `q_cat[i].param_val` are ``ndarrays`` of
    `shape=(nparam_i, nfac_i)`, the Kullback-Leibler divergence is also an
    ``m-tuple`` where `kld[i].param_val` is an ``ndarray`` of `shape=(nparam,)`.


    Notes
    -----
    Note that `t*log(t/a) -> 0` as `t -> 0`. Therefore, we require that `q_i(x)
    = 0` implies `p_i(x) = 0` in which case we can put `kld(x) = 0`. However,
    due to the (inaccurate) numerical representation of the probability maps, it
    is possible that `q_i(x) = 0` and `p_i(x) > 0` (f.e. when `q` has been
    approximated from a relative small set of models). In this case we enforce
    `q_i(x) = p_i(x)` what leads to `kld(x) = 0`.


    Parameters
    ----------
    p_cat : m-tuple
        Tuple of ``CatProb`` instances with `p_cat[i].param_val` being an
        ``ndarray`` of `shape=(nparam_i x nfac_i)`
    q_cat : m-tuple
        Tuple of ``CatProb`` instances with `q_cat[i].param_val` being an
        ``ndarray`` of `shape=(nparam_i, nfac_i)`


    Returns
    -------
    m-tuple
        Tuple of kld maps

        `return[i]` : ``None`` or ``ContParam``
            Return value `i` is ``None`` if `p_cat[i]`` is ``None``, otherwise
            it is an instance of ``ContParam``
    """
    nmtype = len(p_cat)

    # Compute kld only for not-None probability maps
    kld = [None for _ in range(nmtype)]
    cat_mtype = [imtype for imtype in range(nmtype)
                 if not p_cat[imtype] is None]

    # Iteratively compute kld for each mtype
    for imtype in cat_mtype:
        kld[imtype] = ContParam()
        p_i = p_cat[imtype].param_val
        q_i = q_cat[imtype].param_val
        # Check for "corrupted" prior probability map q
        corr_ind = np.where(
            np.logical_and(q_i <= NP_MIN_TOL, p_i > NP_MIN_TOL))[0]
        if corr_ind.size > 0:
            q_i[corr_ind, :] = p_i[corr_ind, :]
            warnings.warn(msg.warn1002(imtype))

        # Compute KLD
        nparam, ncat = p_cat[imtype].param_val.shape
        kld[imtype].param_val = np.zeros((nparam,))
        for icat in range(ncat):
            nnz_ind = p_i[:, icat] > NP_MIN_TOL
            kld[imtype].param_val[nnz_ind]\
                += p_i[nnz_ind, icat] *\
                   np.log(p_i[nnz_ind, icat] / q_i[nnz_ind, icat])

        # Correct (theoretically impossible) negative kld values
        kld[imtype].param_val[np.abs(kld[imtype].param_val) < NP_MIN_TOL] = 0
        if np.any(kld[imtype].param_val < -NP_MIN_TOL):
            ind = np.where(kld[imtype].param_val < -NP_MIN_TOL)[0]
            raise ValueError(msg.err1002(imtype, ind))
    return tuple(kld)


# Hard conditioning data
def generate_hd(popex, meth_w_hd, ncmod, kld, p_cat, q_cat):
    """ `generate_hd` generates the hard conditioning data set that is used
    to sample a new model.

    This set of hard conditioning data does NOT include prior hard conditioning.
    For each model type `(imtype)`, every hard conditioning is obtained by the
    following 2-steps:

        (a) Sample a location `[j]` according to the values in the Kullback-
            Leibler divergence map (i.e. the values in `kld[imtype].param_val`)
        (b) Sample a model `[k]` according to the weights from
            :meth:`compute_w_lik` and directly extract the hard conditioning
            value from `popex.model[k][imtype].param_val[j]`.

    In addition to the hard conditioning, this function also extracts
    probability values from `q_cat` and `p_cat` at the conditioning location.
    These values represent the prior/weighted category probability of the
    category that corresponds to `popex.model[k][imtype].param_val[j]`. They
    can be useful to compute the sampling weight ratio.

    Notes
    -----
    There are two important things to note:

        (1) The two objects `hd_prior` and `hd_generation` are the corresponding
            prior and weighted probability values of the hard conditioning
            CATEGORY that corresponds to the values in `hd_param_val`.
            Therefore, if they are used in the computation of the sampling
            weight ratio, one uses CATEGORY probabilities and NOT value
            probabilities.

        (2) Numerical imperfections (for example in the computation of 'q_cat')
            can cause locations where `p_cat > 0` but `q_cat = 0`. In the
            computation of the Kullback-Leibler divergence we did put
            corresponding `kld` values to `0` (by enforcing `q_i(x) = p_i(x)`)
            and therefore it is impossible to sample and condition such
            locations.


    Parameters
    ----------
    popex : PoPEx
        PoPEx main structure
    meth_w_hd : dict
        Method to compute hard conditioning weights (cf. :meth:`compute_w_lik`)
    ncmod : m-tuple
        Number of conditioning points per model type
    kld : m-tuple
        Tuple of ``ContParam`` instances defining the Kullback-Leibler
        divergence
    p_cat : m-tuple
        Tuple of ``CatProb`` instances defining the weighted category
        probabilities with `p_cat[i].param_val` being an ``ndarray`` of
        `shape=(nparam_i, nfac_i)`
    q_cat : m-tuple
        Tuple of ``CatProb`` instances defining the weighted category
        probabilities with `q_cat[i].param_val` being an ``ndarray`` of
        `shape=(nparam_i x nfac_i)`


    Returns
    -------
    hd_param_ind : m-tuple
        Tuple of hard conditioning indices where hard conditioning values are
        imposed. If there is no hard conditioning for a model type `i`, then
        `hd_ind[i]` is ``None`` otherwise it is an ``ndarray`` of
        `shape=(ncmod[i], ndarray)`.
    hd_param_val : m-tuple
        Tuple of hard conditioning values that are imposed at the hard
        conditioning indices. If there is no hard conditioning for a model type
        `i`, then `hd_val[i]` is ``None`` otherwise it is an ``ndarray`` of
        `shape=(ncmod[i], ndarray)`.
    hd_prior :  m-tuple
        Tuple of probability values according to the prior probability maps in
        `q_cat`. Each value corresponds to the prior probability of the category
        that contains the extracted hard conditioning value.  If there is no
        hard conditioning for a model type `i`, then `hd_val[i]` is ``None``
        otherwise it is an ``ndarray`` of `shape=(ncmod[i], ndarray)`.
    hd_generation : m-tuple
        Tuple of probability values according to the sampling probability maps
        in `p_cat`. Each value corresponds to the sampling probability of the
        category that contains the extracted hard conditioning values. If there
        is no hard conditioning for a model type `i`, then `hd_val[i]` is
        ``None`` otherwise it is an ``ndarray`` of `shape=(ncmod[i], ndarray)`.
    """
    # Raise all numpy errors
    old_settings = np.seterr(all='raise')

    # Initialisation
    nmtype = popex.nmtype
    ncmod = list(ncmod)
    hd_mtype = [imtype for imtype, nc_i in enumerate(ncmod) if nc_i > 0]
    hd_param_ind = [None for _ in range(nmtype)]
    hd_param_val = [None for _ in range(nmtype)]
    hd_prior = [None for _ in range(nmtype)]
    hd_generation = [None for _ in range(nmtype)]

    # STEP (a) (see remark in docstring)
    # Choose hard conditioning locations
    for imtype in hd_mtype:
        # Test KLD values
        if np.any(kld[imtype].param_val < -NP_MIN_TOL):
            ind = np.where(kld[imtype].param_val < -NP_MIN_TOL)[0]
            raise ValueError(msg.err1002(imtype, ind))
        if np.sum(kld[imtype].param_val) == 0:
            ncmod[imtype] = 0

        # Normalize kld values
        try:
            kld_val = kld[imtype].param_val
            nnz_ind = kld[imtype].param_val >= NP_MIN_TOL

            kld_nrm = np.zeros_like(kld_val)
            kld_nrm[nnz_ind] = kld_val[nnz_ind] / np.sum(kld_val[nnz_ind])
        except (ZeroDivisionError, FloatingPointError,
                RuntimeWarning, RuntimeError):
            kld_nrm = np.ones_like(kld[imtype].param_val)\
                       / kld[imtype].nparam
            warnings.warn(msg.warn1001)

        # Compute conditioning location indices
        hd_param_ind[imtype] = \
            np.unique(
                np.random.choice(kld[imtype].nparam,
                                 size=ncmod[imtype],
                                 replace=True,
                                 p=kld_nrm))
        # We put 'replace=True' in order to allow nc_i to be larger than
        # nparam. This can be convenient for the client when nparam is very
        # small. However, multiple indices are removed by using 'np.unique'
        ncmod[imtype] = hd_param_ind[imtype].size

    # Adapt the variables
    hd_mtype = [imtype for imtype, nc_i in enumerate(ncmod) if nc_i > 0]
    hd_param_ind = tuple(hd_param_ind)

    # STEP (b) (see remark in docstring)
    # Compute normalized w_hd
    w_hd_nrm = compute_w_lik(popex=popex, meth=meth_w_hd)

    # Choose hard conditioning values
    for imtype in hd_mtype:
        # Initialize
        hd_param_val[imtype] = []
        hd_prior[imtype] = []
        hd_generation[imtype] = []

        # Select conditional value sequentially
        nc_i = ncmod[imtype]
        if not nc_i == hd_param_ind[imtype].size:
            raise ValueError(msg.err1009)
        for ihd in range(nc_i):
            # Select model according to the 'w_hd' values
            imod = int(np.random.choice(popex.nmod, 1, p=w_hd_nrm))
            hd_ind = hd_param_ind[imtype][ihd]

            # Load pickled model and extract hd value
            path_mod = popex.path_res + popex.model[imod]
            with open(path_mod, 'rb') as mfile:
                model = pickle.load(mfile)
            hd_param_val[imtype].append(model[imtype].param_val[hd_ind])

            # Extract the corresponding category probability value
            icat = model[imtype].param_cat[hd_ind]
            hd_prior[imtype].append(q_cat[imtype].param_val[hd_ind, icat])
            hd_generation[imtype].append(
                p_cat[imtype].param_val[hd_ind, icat])

        # Transform ...[imtype] into ndarray
        hd_param_val[imtype] = np.array(hd_param_val[imtype])
        hd_prior[imtype] = np.array(hd_prior[imtype])
        hd_generation[imtype] = np.array(hd_generation[imtype])

        # Test proability values
        if np.any(hd_prior[imtype] == 0)\
                or np.any(hd_generation[imtype] == 0):
            raise ValueError(msg.err1008)

    # Transform into m-tuples
    hd_param_val = tuple(hd_param_val)
    hd_prior = tuple(hd_prior)
    hd_generation = tuple(hd_generation)

    # Set numpy settings back
    np.seterr(**old_settings)
    return hd_param_ind, hd_param_val, hd_prior, hd_generation


def merge_hd(hd_param_ind_1, hd_param_ind_2, hd_param_val_1, hd_param_val_2):
    """ `merge_hd` used for merging two sets of hard conditioning data.

    It is assumed that `hd_param_ind_i[imtype]` is ``None`` if and only if
    `hd_param_val_i[imtype]` is ``None``.


    Parameters
    ----------
    hd_param_ind_1 : m-tuple
        First set of hard conditioning indices
    hd_param_ind_2 : m-tuple
        Second set of hard conditioning indices
    hd_param_val_1 : m-tuple
        First set of hard conditioning values
    hd_param_val_2 : m-tuple
        Second set of hard conditioning values


    Returns
    -------
    hd_ind : m-tuple
        Merged set of hard conditioning indices
    hd_par : m-tuple
        Merged set of hard conditioning values

    """
    nmtype = len(hd_param_ind_1)

    # Initialisation
    hd_param_ind = [None for _ in range(nmtype)]
    hd_param_val = [None for _ in range(nmtype)]

    # Merge data for each model type
    for imtype in range(nmtype):
        if hd_param_ind_1[imtype] is not None \
                and hd_param_ind_2[imtype] is not None:
            hd_param_ind[imtype] = np.append(hd_param_ind_1[imtype],
                                             hd_param_ind_2[imtype])
            hd_param_val[imtype] = np.append(hd_param_val_1[imtype],
                                             hd_param_val_2[imtype])
        elif hd_param_ind_1[imtype] is None \
                and hd_param_ind_2[imtype] is not None:
            hd_param_ind[imtype] = hd_param_ind_2[imtype]
            hd_param_val[imtype] = hd_param_val_2[imtype]
        elif hd_param_ind_1[imtype] is not None \
                and hd_param_ind_2[imtype] is None:
            hd_param_ind[imtype] = hd_param_ind_1[imtype]
            hd_param_val[imtype] = hd_param_val_1[imtype]
    return tuple(hd_param_ind), tuple(hd_param_val)


def compute_ncmod(popex, meth_w_hd=None):
    """ `compute_ncmod` computes, for each model type, the number of
    conditioning points.


    It is assumed that the number of hard data is restricted model type-wise.
    Therefore, the number of conditioning points is also computed model
    type-wise by sampling from an uniform random variable
    `~U(0, popex.ncmax[imtype])`.


    Notes
    -----
    Note that if the total sum of the likelihood values in `popex.p_lik` is
    zero, `ncmod` is set to zero for each model type.


    Parameters
    ----------
    popex : PoPEx
        PoPEx main structure
    meth_w_hd : dict
        Method to compute hard conditioning weights (cf. :meth:`compute_w_lik`)


    Returns
    -------
    m-tuple
        Number of conditioning points per model type
    """

    # Weights according to 'hd_meth'
    w_hd = compute_w_lik(popex=popex, meth=meth_w_hd)

    # Set all ncmod equal to zero for too small weights
    if np.sum(w_hd) >= NP_MIN_TOL:
        nc_bnd = popex.ncmax
    else:
        nc_bnd = (0,) * popex.nmtype
    ncmod = tuple(np.random.randint(nc_bnd[imtype] + 1, size=1)[0]
                  for imtype in range(popex.nmtype))
    return ncmod


def compute_w_lik(popex, meth=None):
    """ `compute_w_lik` returns the set of normalized likelihood values.

    In practice, when the likelihood values must be represented by a floating
    point number, it might be advantageous to compute approximations of `L(m)`.

    There are several approximation possibilities that are implemented in this
    version (specified in `meth`):

        (a) No approximation (`meth={'name': 'exact'}` or `meth=None`):

            `L(m) = exp( 'log_p_lik' )`

        (b) Sqrt-unskewed (`meth={'name': 'exp_sqrt_log'}`)

            `L(m) ~ exp( -sqrt(-'log_p_lik' )`

        (c) K-unskewed (`meth={'name': 'exp_sqrt_log', 'pow': k}`)

            `L(m) ~ exp( - (-'log_p_lik')^k )`

        (d) Inverse log (`meth={'name': 'inv_log'}`)

            `L(m) ~ 1 / ( 1-'log_p_lik' )`

        (e) Inverse sqrt-log (`meth={'name': 'inv_sqrt_log'}`)

            `L(m) ~ 1 / ( 1+sqrt(-'log_p_lik') )`.

    As mentioned above, these techniques aim to unskew the likelihood values.


    Notes
    -----
    This function is used in two different locations (with possibly two
    different approximation techniques): for the learning
    scheme in the PoPEx sampling and for computing predictions. While in first
    case any approximation technique can be used, the latter choice might bias
    the computation weights.


    Parameters
    ----------
    popex : PoPEx
        PoPEx main structure
    meth : dict
        Defines the approximation method to be used. Fields are

            - ``'name'`` : Name of the method (`str`)
            - ``'pow'`` : Power for method (c) (`float`)


    Returns
    -------
    ndarray, shape=(nmod,)
        Array of normalized weights

    """

    log_w_lik = np.array(popex.log_p_lik)
    if log_w_lik.size == 0:
        return np.zeros_like(log_w_lik)

    if meth is None:
        name = 'exact'
    else:
        name = meth['name']

    # Weight quantities according to 'p_lik'
    if name == 'exact':
        # Shift log-weights (possible because of the normalization at the end)
        log_w_lik -= np.max(log_w_lik)

        # Compute exp for sufficiently large log-values
        w_lik = np.zeros_like(log_w_lik)
        nnz_ind = log_w_lik >= np.log(NP_MIN_TOL)
        w_lik[nnz_ind] = np.exp(log_w_lik[nnz_ind])

    elif name == 'exp_sqrt_log':
        # Shift log-weights (possible because of the normalization at the end)
        log_w_lik -= np.max(log_w_lik)

        # Bring the weights "closer together"
        log_w_lik = -np.sqrt(-log_w_lik)

        # Compute exp for sufficiently large log-values
        w_lik = np.zeros_like(log_w_lik)
        nnz_ind = log_w_lik >= np.log(NP_MIN_TOL)
        w_lik[nnz_ind] = np.exp(log_w_lik[nnz_ind])

    elif name == 'exp_pow_log':
        # Shift log-weights (possible because of the normalization at the end)
        log_w_lik -= np.max(log_w_lik)

        # Bring the weights "closer together"
        log_w_lik = -((-log_w_lik)**(meth['pow']))

        # Compute exp for sufficiently large log-values
        w_lik = np.zeros_like(log_w_lik)
        nnz_ind = log_w_lik >= np.log(NP_MIN_TOL)
        w_lik[nnz_ind] = np.exp(log_w_lik[nnz_ind])

    # Weight quantities according to 'log_p_lik'
    elif name == 'inv_log':
        # Check for all-negative log(w)
        pos_ind = log_w_lik >= NP_MIN_TOL
        if np.any(pos_ind):
            ind = np.where(pos_ind)
            raise ValueError(msg.err1010(ind))

        # Compute 1/(1-log(w))
        w_lik = 1 / (1-log_w_lik)

    # Weight quantities according to sqrt(-'log_p_lik')
    elif name == 'inv_sqrt_log':
        # Check for all-negative log(w)
        pos_ind = log_w_lik >= NP_MIN_TOL
        if np.any(pos_ind):
            ind = np.where(pos_ind)
            raise ValueError(msg.err1010(ind))

        # Compute 1/( 1-log( abs(w)^1/2 ) )
        w_lik = 1 / (1+np.sqrt(-log_w_lik))

    else:
        raise ValueError('\nUnknown method "{}" for computing weights'
                         .format(meth))

    # Remove numerical inaccuracy
    w_lik[w_lik < NP_MIN_TOL] = 0

    # Check weights for non-negativity
    if np.any(w_lik < -NP_MIN_TOL):
        ind = np.where(w_lik < -NP_MIN_TOL)[0]
        raise ValueError(msg.err1003(ind))

    # Weights must be normalized, because
    #   - of the shift when using meth['name'] = 'exact' (see above)
    #   - 'utils.generate_hd' assumes normalized weights
    try:
        nnz_ind = w_lik >= NP_MIN_TOL
        w_lik_nrm = np.zeros_like(w_lik)
        w_lik_nrm[nnz_ind] = w_lik[nnz_ind] / np.sum(w_lik[nnz_ind])
    except (ZeroDivisionError, FloatingPointError,
            RuntimeWarning, RuntimeError):
        w_lik_nrm = np.ones_like(popex.log_p_lik) / popex.nmod
        warnings.warn(msg.warn1004)

    return w_lik_nrm


# Generic functions
def compute_w_pred(popex, nw_min=0, ibnd=-1, meth=None):
    """ `compute_w_pred` returns the set of normalized predictive weights.

    For assuring a minimum number of effective weights, they are computed such
    that

        `ne(w_pred) = min(nw_min, ne(w))`

    where `w` contains the weights associated to the models and `ne(w)` denotes
    the number of effective weights. This quantity can be modified by replacing
    `w` with `w^\alpha`, where `alpha > 0`. A `1-d` optimisation problem is used
    to compute the optimal `\alpha` value.


    Parameters
    ----------
    popex : PoPEx
        PoPEx main structure
    nw_min : int
        Mininum number of effective weights `(= l_0)`
    ibnd : int
        Length of the weight array
    meth : dict
        Defines the approximation method to be used (cf. :meth:`compute_w_lik`).
        Fields are

            - ``'name'`` : Name of the method (`str`)
            - ``'pow'`` : Power for method (c) (`float`)


    Returns
    -------
    ndarray, shape=(nmod,)
        Array of predictive weights

    """

    # Compute weight array
    w_lik = compute_w_lik(popex=popex, meth=meth)
    ratio = np.exp(popex.log_p_pri-popex.log_p_gen)
    weights = w_lik * ratio

    # Adapt the weights by the power method
    if ibnd > 0:
        weights = weights[:ibnd]

    if np.sum(weights > NP_MIN_TOL) > 0:
        ne_w = isampl.ne(weights)
        # ne_w_pred = nw_min + ne_w
        ne_w_pred = np.max((nw_min, ne_w))
        w_pred = isampl.correct_w(weights, ne_w_pred)
    else:
        w_pred = np.ones_like(weights) / weights.size

    return w_pred


def compute_subset_ind(p_frac, weights):
    """ `compute_subset_ind` computes the smallest index set that covers a given
    percentage.

    This means that the subset indices `ind` are such that

        `np.sum(weights[ind]) >= p_frac * np.sum(weights)`,

    or in other words `weights[ind]` covers at least a fraction of `p_frac` of
    the total some of `weights`.


    Parameters
    ----------
    p_frac : float
        Coverage fraction in `(0, 1]`
    weights : ndarray, shape=(nw,)
        Non-negative weights


    Returns
    -------
    list
        Subset of indices

    """
    # Compute cumsum with largest weights first
    w_sorted = np.sort(weights)[::-1]
    w_cs = np.cumsum(w_sorted)

    # Compute treshold weight value (for p_frac)
    val = w_sorted[np.where(w_cs / w_cs[-1] - p_frac >= -NP_MIN_TOL)[0][0]]

    return [ind for ind, w in enumerate(weights) if w >= val]


def write_hd_info(popex, imod, hd_param_ind, hd_param_val):
    """ `write_hd_info` writes the hard conditioning that has been deduced
    for creating a specific model to a text file.

    | The text file is saved at `popex.path_res` with the following structure:
    |   <popex.path_res>$
    |       └-- hd
    |           └-- hd_modXXXXXX.txt


    Parameters
    ----------
    popex : PoPEx
        PoPEx main structure
    imod : int
        Model index
    hd_param_ind : m-tuple
        Hard conditioning indices
    hd_param_val : m-tuple
        Hard conditioning values


    Returns
    -------
    None

    """
    # Path of the result location
    path_res = popex.path_res
    nmtype = popex.nmtype

    # Make hd directory
    path_hd = path_res + 'hd/'
    try:
        os.mkdir(path_hd)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
        pass

    # For each model, write the precise hard conditioning
    with open('{}hd_mod{:06}.txt'.format(path_hd, imod), 'w+') as file:
        for imtype in range(nmtype):
            file.write('mtype {:02}\n'.format(imtype) + '-'*8 + '\n')
            if hd_param_ind[imtype] is not None:
                nc_i = hd_param_ind[imtype].size
                file.write('nc = {:3d}\n'.format(nc_i))
                file.write('Index   Value\n')
                mat = np.concatenate((hd_param_ind[imtype].reshape(-1, 1),
                                      hd_param_val[imtype].reshape(-1, 1)),
                                     axis=1)
                for li in mat:
                    val, pos = str(li)[1:-1].split()
                    file.write(' {:<7} {}'.format(val, pos) + '\n')
            else:
                file.write('None\n')
            file.write('\n')
    return None


def write_run_info(pb, popex, imod, log_p_lik, cmp_log_p_lik,
                   log_p_pri, log_p_gen, ncmod):
    """ `write_run_info` writes some algorithm specific information to a text
    file.

    | The text file is save at `popex.path_res` with the following structure:
    |   <popex.path_res>$
    |       └-- run_info.txt


    Parameters
    ----------
    pb : Problem
        Defines the problem functions and parameters
    popex : PoPEx
        PoPEx main structure
    imod : int
        Model index
    log_p_lik : float
        Log-likelihood value of the model
    cmp_log_p_lik : bool
        Indicates if likelihood has been computed (True) or predicted (False)
    log_p_pri : float
        Prior log-probability of the model
    log_p_gen : float
        Sampling log-probability of the model
    ncmod : m-tuple
        Model type specific number of conditioning points used


    Returns
    -------
    None

    """
    # Path of the result location
    path_res = popex.path_res
    nmtype = popex.nmtype

    # Compute n_e diagnostics
    with warnings.catch_warnings(record=True) as _:
        ne_l    = isampl.ne(compute_w_lik(popex=popex,
                                          meth={'name': 'exact'}))
        ne_w_hd = isampl.ne(compute_w_lik(popex=popex,
                                          meth=pb.meth_w_hd))
        ne_w_l  = isampl.ne(compute_w_pred(popex=popex,
                                           nw_min=0,
                                           meth={'name': 'exact'}))

    # Write into a file 'iteration_info.txt'
    if write_run_info.__first_call:
        mode = 'w+'
        write_run_info.__first_call = False
    else:
        mode = 'a'
    with open('{}/run_info.txt'.format(path_res), mode) as file:

        # Write model details
        file.write('Model {:6d}\n'.format(imod) + '-' * 12 + '\n')
        file.write('\n')
        file.write('  Hard conditioning\n')
        for imtype in range(nmtype):
            file.write('    mtype {:2d}:  nc = {:3d}\n'
                       .format(imtype, ncmod[imtype]))
        file.write('\n')
        file.write('  Computations\n')
        file.write('    log(p_lik)  = {:17.7f} (cmp = {!r})\n'
                   .format(log_p_lik, cmp_log_p_lik))
        file.write('    log(p_pri)  = {:17.7f}\n'.format(log_p_pri))
        file.write('    log(p_gen)  = {:17.7f}\n'.format(log_p_gen))
        file.write('    weight      = {:17.7f}\n'
                   .format(np.exp(log_p_lik + log_p_pri - log_p_gen)))

        # Write n_e
        file.write('\n')
        file.write('  Diagnostics\n')
        file.write('    ne(L)       = {:11.1f}\n'.format(ne_l))
        file.write('    ne(w_hd)    = {:11.1f}\n'.format(ne_w_hd))
        file.write('    ne(w_pred)  = {:11.1f}\n'.format(ne_w_l))
        file.write('\n\n')
write_run_info.__first_call = True



# -*- coding: utf-8 -*-
""" `popex_objects.py` contains the PoPEx-specific class definitions.

Main structure
    - :class:`PoPEx`: Main class for any PoPEx simulation. It contains the
      model chain and any corresponding probability measures.

Sampling definitions
    - :class:`Problem`: Defines the sampling parameters and functions
    - :class:`Learning`: Learning scheme for learning the likelihood values
    - :class:`Prediction`: Defines the prediction parameters and functions

Classes associated to a model type
    - :class:`MType`: (`abstract`) Parent class for each map associated to a
      model type
    - :class:`ContParam`: (inherits from :class:`MType`) Class for each map that
      is associated to the model types but not to categories (e.g.
      `kld[imtype]`, `entropy[imtype]`)
    - :class:`CatMType`: (`abstract`, inherits from :class:`MType`) Parent class
      for each map that is associated to categories
    - :class:`CatProb`: (inherits from :class:`CatMType`) This class is used for
      the representation of probability distributions over categories
      (e.g. `p_cat[imtype]`, `q_cat[imtype]`)
    - :class:`CatParam`: (inherits from :class:`CatMType`) This class is used
      for the representation of categorized parameter values (e.g.
      `model[j][imtype]`)

"""

# -------------------------------------------------------------------------
#   Authors: Christoph Jaeggli, Julien Straubhaar and Philippe Renard
#   Year: 2019
#   Institut: University of Neuchatel
#
#   Copyright (c) 2019 Christoph Jaeggli
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

# General imports
import abc
import numpy as np
import reprlib
from copy import deepcopy
import pickle

# Package imports
import popex.popex_messages as msg


# Main structure
class PoPEx:
    """ Main class for any PoPEx simulation.

    This class is the main object for the PoPEx algorithm. It contains all
    the models, likelihood, log-prior and log-generation information of a PoPEx
    run.

    Parameters
    ----------
    model : list
        List of models

        model[j] : m-tuple
            Tuple of ``MType`` instances
    log_p_lik : ndarray, shape=(nmod,)
        Natural logarithm of likelihood measure
    cmp_log_p_lik : ndarray, shape=(nmod,)
        Boolean indicator whether the log-likelihood value has been computed
        (True) or predicted (False)
    log_p_pri : ndarray, shape=(nmod,)
        Natural logarithm of prior measure value
    log_p_gen : ndarray, shape=(nmod,)
        Natural logarithm of sampling measure value
    ncmax : m-tuple
        Maximum number of conditioning points for each model type
    nc : list
        List of m-tuples

        nc[j] : m-tuple
            Contains the number of conditioning used conditioning points in the
            generation of `model[j]`
    nmtype : int
        Number of model types
    path_res : str
        Path of the results

    """

    def __init__(self,
                 model=None,
                 log_p_lik=np.empty(0),     # Don't change this, because
                                            # sum(...) must be zero initially
                                            # this ensures that we take zero
                                            # hard conditioning for the first
                                            # models
                 cmp_log_p_lik=np.empty(0, dtype=bool),
                 log_p_pri=np.empty(0),
                 log_p_gen=np.empty(0),
                 ncmax=(0,),
                 nc=None,
                 nmtype=1,
                 path_res='~/'):

        # Initialisation of an instance
        if model is None:
            self.model = []
        else:
            self.model = deepcopy(model)
        self.log_p_lik = deepcopy(log_p_lik)
        self.cmp_log_p_lik = deepcopy(cmp_log_p_lik)
        self.log_p_pri = deepcopy(log_p_pri)
        self.log_p_gen = deepcopy(log_p_gen)
        self.ncmax = ncmax
        if nc is None:
            self.nc = []
        else:
            self.nc = deepcopy(nc)
        self.nmtype = nmtype
        self.path_res = path_res

    def __repr__(self):
        rlib = reprlib.Repr()
        rlib.maxlist = 3

        # Define repr string
        class_name = type(self).__name__
        model_str = rlib.repr(self.model)
        log_p_lik_str = rlib.repr(self.log_p_lik)
        cmp_log_p_lik_str = rlib.repr(self.cmp_log_p_lik)
        log_p_pri_str = rlib.repr(self.log_p_pri)
        log_p_gen_str = rlib.repr(self.log_p_gen)
        nc_str = rlib.repr(self.nc)
        return '{}(model={},\nlog_p_lik={},\ncmp_log_p_lik={},\n' \
               'log_p_pri={},\nlog_p_gen={},\nncmax={!r},\nnc={},\n' \
               'nmtype={},\npath_res={})' \
               .format(class_name,
                       model_str,
                       log_p_lik_str,
                       cmp_log_p_lik_str,
                       log_p_pri_str,
                       log_p_gen_str,
                       self.ncmax,
                       nc_str,
                       self.nmtype,
                       self.path_res)

    def __str__(self):
        return self.__repr__()

    def add_model(self, imod, model, log_p_lik, cmp_log_p_lik,
                  log_p_pri, log_p_gen, ncmod):
        """ Appends a new model at the end of the model list and updates the
        measure arrays.


        Parameters
        ----------
        imod : int
            Model index
        model : m-tuple
            Tuple of ``MType`` instances defining a new model
        log_p_lik : float
            Log-likelihood value of model
        cmp_log_p_lik : bool
            Indicates if the log-likelihood has been computed (True) or
            predicted (False)
        log_p_pri : float
            Log-prior value of model
        log_p_gen : float
            Log-generation value of model
        ncmod : m-tuple
            Defines the number of conditioning points that have been used in the
            generation of the model


        Returns
        -------
        None

        """
        # Pickle the model to save memory
        path_mod = 'model/mod{:06d}.mod'.format(imod)
        with open(self.path_res + path_mod, 'wb') as mfile:
            pickle.dump(model, mfile)

        # Add model information to self
        self.model.append(path_mod)
        self.log_p_lik     = np.append(self.log_p_lik, log_p_lik)
        self.cmp_log_p_lik = np.append(self.cmp_log_p_lik, cmp_log_p_lik)
        self.log_p_pri     = np.append(self.log_p_pri, log_p_pri)
        self.log_p_gen     = np.append(self.log_p_gen, log_p_gen)
        self.nc.append(ncmod)

    def insert_model(self, loc, imod, model, log_p_lik, cmp_log_p_lik,
                     log_p_pri, log_p_gen, ncmod):
        """ Inserts a new model at `loc` of the model list and  updates the
        measure arrays.


        Parameters
        ----------
        loc :  int
            Location of the insertion
        imod : int
            Model index
        model : m-tuple
            Tuple of ``MType`` instances defining a model
        log_p_lik : float
            Log-likelihood value of model
        cmp_log_p_lik : bool
            Indicates if the log-likelihood has been computed (True) or
            predicted (False)
        log_p_pri : float
            Log-prior value of model
        log_p_gen : float
            Log-generation value of model
        ncmod : m-tuple)
            Defines the number of conditioning points that have been used in the
            generation of the model


        Returns
        -------
        None

        """
        # Pickle the model to save memory
        path_mod = 'model/mod{:06d}.mod'.format(imod)
        with open(self.path_res + path_mod, 'wb') as mfile:
            pickle.dump(model, mfile)

        # Insert the model and the values at 'loc'
        self.model.insert(loc, path_mod)
        self.log_p_lik     = np.insert(self.log_p_lik,     loc, log_p_lik)
        self.cmp_log_p_lik = np.insert(self.cmp_log_p_lik, loc, cmp_log_p_lik)
        self.log_p_pri     = np.insert(self.log_p_pri, loc, log_p_pri)
        self.log_p_gen     = np.insert(self.log_p_gen, loc, log_p_gen)
        self.nc.insert(loc, ncmod)

    @property
    def nmod(self):
        """ Number of models

        Returns
        -------
        int
            Number of models in `model`

        """
        return len(self.model)


# Problem definitions
class Problem:
    """ Defines the sampling problem that should be addressed by the PoPEx
    method.

    The user must provide function definitions for `generate_m` and
    `compute_log_p_lik`. For the definition of the model space, we can also
    provide 'prior hard conditioning' through the function `get_hd_pri`.
    Optionally, a likelihood learning scheme can be defined in
    `learning_scheme`. Furthermore, one also must define how to compute the
    ratio in the importance sampling weights. For this, the functions
    `compute_log_p_pri` and `compute_log_p_gen` can also be defined manually. If
    they are left empty, the default version that only considers the hard
    conditioning data points is used.


    Parameters
    ----------
    generate_m : function
        Generates a new model m from a set of hard conditioning data.

        `generate_m(hd_param_ind, hd_param_val, imod)`

        Parameters:
            - hd_param_ind : m-tuple
                For each instance in the model tuple, this variable defines the
                hard conditioning INDICES (where to apply HD). `hd_param_ind[i]`
                is an ``ndarray`` of `shape=(nhd_i,)`
            - hd_param_val : m-tuple
                For each instance in the model tuple, this variable defines the
                hard conditioning VALUES (what to imposed). `hd_param_val[i]` is
                an ``ndarray`` of `shape=(nhd_i,)`
            - imod : int
                Model index

        Returns:
            m-tuple
                New model such as `(CatParam_1, ..., CatParam_m)`
    compute_log_p_lik : function
        Computes the natural logarithm of the likelihood of a model. It usually
        runs an expensive forward operation and compares the response to a given
        set of observations.

        `compute_log_p_lik(model, imod)`

        Parameters:
            - model : m-tuple
                Model such as `(CatParam_1, ..., CatParam_m)` (cf. output of
                `generate_m()`)
            - imod : int
                Model index

        Returns:
            float
                Log-likelihood value of the model
    get_hd_pri : function
        Provides the 'prior hard conditioning' that is used in the definition of
        the model space (i.e. parameter values that are known without
        uncertainty).

        `get_hd_pri()`

        Returns
            - hd_pri_ind : m-tuple
                For each instance in the model tuple, this variable defines the
                hard conditioning INDICES.
            - hd_pri_val : m-tuple
                For each instance in the model tuple, this variable defines the
                hard conditioning VALUES.
    compute_log_p_pri : function, optional
        This function computes the log-prior probability of a model that has
        been generated from a given set of hard conditioning data. Note that
        it's definition is OPTIONAL. If it is left undefined, a default
        implementation will be used (see remark below).

        `compute_log_p_pri(model, hd_p_pri, hd_param_ind)`

        Parameters:
            - model : m-tuple
                Model such as `(CatParam_1, ..., CatParam_m)` (cf. output of
                `generate_m()`)
            - hd_p_pri : m-tuple
                Tuple of the hard conditioning probability values for a given
                model. Each probability value describes the prior probability of
                observing the category of the model value at the corresponding
                conditioning location.
                hd_p_pri[i] is an ``ndarray`` of `shape=(nhd_i,)`
            - hd_param_ind : m-tuple
                Hard conditioning indices
        Returns:
            float
                Log-prior probability value

    compute_log_p_gen : function, optional
        This function computes the log-probability of generating a model in the
        PoPEx sampling from a given set of hard conditioning. Note that it's
        definition is OPTIONAL. If it is left undefined, a default
        implementation will be used (see remark below).

        `compute_log_p_gen(model, hd_p_gen, hd_param_ind)`

        Parameters:
            - model : m-tuple
                Model such as `(CatParam_1, ..., CatParam_m)` (cf. output of
                `generate_m()`)
            - hd_p_gen : m-tuple
                Tuple of the hard conditioning probability values for a given
                model. Each probability value describes the prior probability of
                observing the category of the model value at the corresponding
                conditioning location. `hd_p_gen[i]` is an ``ndarray`` of
                `shape=(nhd_i,)`
            - hd_param_ind : m-tuple
                Hard conditioning indices
        Returns:
            float
                Log-generation probability value
    learning_scheme : Learning, optional
        Learning scheme for log_p_lik (concrete sublcass of ``Learning``)
    meth_w_hd : dict, optional
        Defines the method for computing the learning weights that are used in
        the computation of the hard conditioning points (cf.
        :meth:`compute_w_lik`)
    nmtype : int
        Number of model types
    seed : int
        Initial seed


    Notes
    -----

    (1) Let us provide a simple example for hard conditioning data in
        `hd_param_ind` and `hd_param_val`. It is important to note that PoPEx
        does NOT use any parameter locations. They might be defined by the
        user. If so, they have to follow a certain structure. Let the parameter
        locations be such that::

                                        x    y    z
            param_loc[0] = np.array([[0.5, 1.5, 0.5],   # Parameter 0
                                     [0.5, 2.5, 0.5],   # Parameter 1
                                     [0.5, 3.5, 0.5]]   # Parameter 2

        and the parameter indices (in `hd_param_ind`) are for example::

            hd_param_ind[0] = [0, 2]    # Condition parameter 0 and 2

        so we will use `param_loc[0][hd_param_ind[0], :]` for obtaining the
        array::

            np.array([[0.5, 1.5, 0.5],
                      [0.5, 3.5, 0.5]]).

        This array indicates the physical locations where hard conditioning
        should be applied for the model type `0`. Let the parameter values
        (in `hd_param_val`) be given by::

            hd_param_val[0] = np.array([1.2, 2.5]).

        Together with the conditioning locations above, this imposes hard
        conditioning data as follows::

             x     y     z     val
            0.5   1.5   0.5    1.2
            0.5   3.5   0.5    2.5

    (2) Note that it is possible to NOT define `compute_log_p_pri` and
        `compute_log_p_gen`. In this case, a predefined function will be used.
        This predefined implementation assumes that the quantities `p_pri` and
        `p_gen` are only used TOGETHER in the form of a RATIO

            `ratio(m) = rho(m) / phi(m)`.

        In other words, the default functions assume that we are only interested
        in the DIFFERENCE of the log values, i.e.

            `log_p_pri - log_p_gen`,

        and never in the exact values on their own. It is left to the user to
        implement a more suitable computation, whenever the above assumption is
        not sufficient. For more information also consult the theoretical
        description of the PoPEx method.
    (3) It is also possible to NOT define the `learning_scheme`. In this case,
        the log-likelihood value will ALWAYS be computed.

    """

    def __init__(self,
                 generate_m,
                 compute_log_p_lik,
                 get_hd_pri,
                 compute_log_p_pri=None,
                 compute_log_p_gen=None,
                 learning_scheme=None,
                 meth_w_hd=None,
                 nmtype=1,
                 seed=0):

        # Mandatory arguments
        self.generate_m = generate_m
        self.compute_log_p_lik = compute_log_p_lik
        self.get_hd_pri = get_hd_pri

        # Optional arguments
        if compute_log_p_pri is None:
            self.compute_log_p_pri = _default_log_p
        else:
            self.compute_log_p_pri = compute_log_p_pri
        if compute_log_p_gen is None:
            self.compute_log_p_gen = _default_log_p
        else:
            self.compute_log_p_gen = compute_log_p_gen
        self.learning_scheme = learning_scheme

        # Method
        if meth_w_hd is None:
            self.meth_w_hd = {'name': 'exact'}
        else:
            self.meth_w_hd = dict(meth_w_hd)

        # Constants
        self.nmtype    = int(nmtype)
        self.seed      = int(seed)

    def __repr__(self):
        class_name = type(self).__name__
        return '{}(generate_m={},\n' \
               'compute_log_p_lik={},\n' \
               'get_hd_pri={},\n' \
               'compute_log_p_pri={},\n' \
               'compute_log_p_gen={},\n' \
               'learning_scheme={!r},\n' \
               'meth_w_hd={!r},\n' \
               'nmtype={},\n' \
               'seed={})'\
               .format(class_name,
                       self.generate_m.__name__,
                       self.compute_log_p_lik.__name__,
                       self.get_hd_pri.__name__,
                       self.compute_log_p_pri.__name__,
                       self.compute_log_p_gen.__name__,
                       self.learning_scheme,
                       self.meth_w_hd,
                       self.nmtype,
                       self.seed)

    def __str__(self):
        return self.__repr__()


def _default_log_p(model, hd_prob, hd_ind):
    """ `_default_log_p` is a default definition of the function that computes
    the log-probability of sampling a model from a given set of hard
    conditioning data. In this definition, it is assumed that the hard
    conditioning locations and all the model types are INDEPENDENT. From these
    assumptions, it follows that the probability of sampling a model is obtained
    by

        `p = p_1 * ... * p_nmtype`

    where

        `p_i = hd_prob[i][1] * ... * hd_prob[i][nc_i]`.

    Therefore, the log-probability is then given by

        `log_p_gen = sum_i sum_j log(hd_prob[i][j])`

    where `i` runs through the model types and `j` through the conditioning
    points within a given model type.


    Parameters
    ----------
    model : m-tuple
        Tuple of 'MType' instances defining a model
    hd_prob : m-tuple
        Tuple of the hard conditioning probabilities for a given model. Each
        probability value quantifies the probability of observing the
        corresponding parameter category at the this location.

            hd_prob[i] : ndarray, shape=(nc_i,)
                Containing the category probability values
    hd_ind : m-tuple
        Hard conditioning indices (for a more detailed explanation see comments
        in 'generate_m')


    Returns
    -------
    float
        Log-probability of sampling a model

    """
    # Raise all numpy errors for this function
    old_settings = np.seterr(all='raise')

    # Compute default log probability
    log_val = 0.
    for hd_prob_i in hd_prob:
        if hd_prob_i is not None:
            try:
                log_val += np.sum(np.log(hd_prob_i))
            except FloatingPointError as err:
                print(hd_prob_i)
                raise FloatingPointError(err)

    # Set numpy settings back
    np.seterr(**old_settings)
    return log_val


class Learning(abc.ABC):
    """ `Learning` defines an abstract parent class for a learning scheme.

    Let's assume that we want to define a learning scheme that predicts the
    log-likelihood of a model. In this case we define an explicite sub-class
    of ``Learning`` and provide implementations of the methods

        - :meth:`train`
        - :meth:`compute_p_eval_for`
        - :meth:`learn_value_of`

    It is assumed that there is a choice between 'evaluating the exact
    answer' (which is very expensive) or 'predicting the  answer by a machine
    learning scheme' (which should be very fast). The learning scheme undergoes
    the following main steps

        - Update the learning scheme regularly by using the function
          :meth:`trian` (cf. `upd_ls_freq` in `algorithm.run_popex_mp`). Note
          that here you can choose to only use likelihood values that have
          effectively been computed (cf. `PoPEx.cmp_log_p_lik`).
        - For a given instance compute a probability `p in [0,1]` with which
          the log-likelihood is predicted or evaluated (cf.
          :meth:`compute_p_eval_for`) and then the value eventually is predicted
          (cf. :meth:`learn_value_of`).

    Notes
    -----
        In the PoPEx framework this can be used to learn the log-likelihood
        values for each model (=value of interest). In this regard, predicting a
        value rather than computing it can considerably improve the overall
        computational time.
    """

    @abc.abstractmethod
    def train(self, popex):
        """ This method creates a learning scheme.

        The learning scheme can be saved as class parameter.

        Parameters
        ----------
        popex : PoPEx
            PoPEx main structure (cf `popex_objects.PoPEx`)


        Returns
        -------
        None

        """

    @abc.abstractmethod
    def compute_p_eval_for(self, model):
        """ Computes and return a probability value in [0, 1] that determines
        whether a model should be evaluated exactly.


        The two extreme confidence values signify:

            `p=0`: Value can be learned from the learning scheme
            `p=1`: Value should be evaluated exactly.


        Parameters
        ----------
        model : m-tuple
            Tuple of ``Mtype`` instances that define the new model


        Returns
        -------
        float
            Probability value in `[0, 1]`

        """

    @abc.abstractmethod
    def learn_value_of(self, model):
        """ Uses the existing learning scheme to learn the value of interest for
        an instance.


        Notes
        -----
        This function should raise an error if there is no existing learning
        scheme.


        Parameters
        ----------
        model : m-tuple
            Tuple of ``Mtype`` instances that define the new model


        Returns
        -------
        float
            Predicted log-likelihood value

        """


class Prediction:
    """ Defines a prediction that should be computed based on an existing
    PoPEx instance.

    The user must provide function definitions for `compute_pred` that actually
    implements the prediction operator. Note that there is no return value
    expected from that function. Any important result can be saved under

        | <path_res>$
        |  └-- solution
        |      └-- pred_<name>_modXXXXXX


    Parameters
    ----------
    compute_pred : function
        Computes the prediction for a given model.

        `compute_pred(popex, imod)`

        Parameters:
            - popex : PoPEx
                PoPEx main structure (cf :class:`popex.popex_objects.PoPEx`)
            - imod : int
                Model index

        Returns:
            None
    meth_w_pred : dict
        Defines the method used for computing the prediction weights (cf.
        :meth:`popex.utils.compute_w_lik`)
    nw_min : float
        Minimum number of effective weights (= l_0)
    wfrac_pred : float
        Number in (0,1] defining the fraction of the total weight to be used for
        the prediction. If `p=1`, all predictions for any model that has
        non-zero weight is computed. If `p<1` we take the minimum number of
        weight to cover a ratio of `p` of the total sum of weights.
    """

    def __init__(self,
                 compute_pred=None,
                 meth_w_pred=None,
                 nw_min=None,
                 wfrac_pred=1.):
        self.compute_pred = compute_pred

        if meth_w_pred is None:
            self.meth_w_pred = {'name': 'exact'}
        else:
            self.meth_w_pred  = dict(meth_w_pred)
        self.nw_min     = float(nw_min)
        self.wfrac_pred = float(wfrac_pred)

    def __repr__(self):
        class_name = type(self).__name__
        return '{}(compute_pred={},\n' \
               'meth_w_pred={},\n' \
               'nw_min={},\n' \
               'wfrac_pred={})' \
            .format(class_name,
                    self.compute_pred.__name__,
                    self.meth_w_pred,
                    self.nw_min,
                    self.wfrac_pred)

    def __str__(self):
        return self.__repr__()


# Classes associated to a model type
class MType(abc.ABC):
    """ This class is the parent of any quantity associated to a model type.

    Parameters
    ----------
    dtype_val : str
        Type of the ``ndarray`` values (eg. 'int8', 'float32', 'float64', etc)
    param_val : ndarray
        Values associated to the parameters
    """

    def __init__(self, dtype_val='float64', param_val=None):
        # Assuring instance attributes 'param_val'
        if param_val is None:
            self.param_val = np.empty((0,), dtype=dtype_val)
        else:
            self.param_val = np.array(param_val, dtype=dtype_val)

    def __repr__(self):
        rlib = reprlib.Repr()
        rlib.maxlist = 3

        class_name = type(self).__name__
        param_val_str = rlib.repr(self.param_val)
        return '{}({!r}, {})'.format(class_name,
                                     self.param_val.dtype.__str__(),
                                     param_val_str)

    def __str__(self):
        class_name = type(self).__name__
        return '\n{}\n   dtype_val = {!r}\n   param_val = ({} ndarray)\n' \
            .format(class_name,
                    self.param_val.dtype.__str__(),
                    self.param_val.shape)

    @abc.abstractmethod
    def __setattr__(self, key, value):
        """ This method must be implemented in each subclass. The main idea is
        to allow updates for other attributes as well as a dimensionality
        checks.
        """

    @property
    def nparam(self):
        """ Number of parameters.

        Returns
        -------
        int
            Number of values in `param_val`
        """
        return self.param_val.shape[0]


class CatMType(MType):
    """ This class is the abstract parent of any quantity associated to a
    categorical model type.

    Parameters
    ----------
    dtype_val : str
        Type of the ``ndarray`` values (eg. 'int8', 'float32', 'float64', etc)
    param_val : ndarray
        Values associated to the parameters
    categories : list
        List of size `ncat`. Each instance of the list is again a list
        of 2-tuples that define the value range for the category.

        If `categories[i] = [(v_1, v_2), (v_3, v_4)]`, where `v_j` are real
        values, then the category `i` is defined by the union

            `[v_1, v_2) U [v_3, v_4)`

    """

    def __init__(self, dtype_val='float64', param_val=None, categories=None):
        # Assuring instance attributes 'param_val' and 'categories'
        if categories is None:
            self.categories = []
        else:
            # To save memory no copy is produced
            self.categories = categories
        super().__init__(dtype_val, param_val)

    def __repr__(self):
        rlib = reprlib.Repr()
        rlib.maxlist = 3

        class_name = type(self).__name__
        param_val_str = rlib.repr(self.param_val)
        cat_str = rlib.repr(self.categories)
        return '{}({!r}\n,{},\n{})'.format(class_name,
                                           self.param_val.dtype.__str__(),
                                           param_val_str,
                                           cat_str)

    def __str__(self):
        class_name = type(self).__name__
        return '\n{}\n   dtype_val = {!r}\n   categories = {!r}\n' \
               '   param_val = ({} ndarray)\n'\
            .format(class_name,
                    self.param_val.dtype.__str__(),
                    '  |  '.join(
                        ' U '.join(
                            '[{}, {})'.format(self.categories[icat][intvl][0],
                                              self.categories[icat][intvl][1])
                            for intvl in range(len(self.categories[icat])))
                        for icat in range(len(self.categories))),
                    self.param_val.shape)

    @abc.abstractmethod
    def __setattr__(self, key, value):
        """ This method must be implemented in each subclass. The main idea is
        to allow updates for other attributes as well as a dimensionality
        checks.
        """

    @property
    def ncat(self):
        """ Number of categories.

        Returns
        -------
        int
            Number of categories in `categories`
        """
        return len(self.categories)


class ContParam(MType):
    """ This class is used to define a map of continuous values where each value
    is associated to a model parameter (e.g. `entropy`, `kld`, etc.).


    Notes
    -----
        The shape of `param_val` is `(nparam,)`.
    """

    def __setattr__(self, key, value):
        """ The '__setattr__' method is customized for a check of
        dimensionality. In 'ContParam' the attribute 'param_val' must be a
        1-dimensional object compatible to an ndarray.
        """
        if key == 'param_val':
            if not isinstance(value, np.ndarray):
                value = np.array(value)
            if len(value.shape) > 1:
                raise AttributeError(msg.err2002)
        self.__dict__[key] = value


class CatProb(CatMType):
    """ This class is used to define a map of continuous values for each
    category, where each value is associated to a model parameter (e.g. `p_cat`,
    `q_cat`, etc.).


    Notes
    -----
        The shape of `param_val` is `shape=(nparam, ncat)`.
    """

    def __setattr__(self, key, value):
        """ The '__setattr__' method is customized for a check of
        dimensionality. In 'CatProb' the attribute 'param_val' must be a
        2-dimensional object compatible to an ndarray.
        """
        if key == 'param_val':
            if not isinstance(value, np.ndarray):
                value = np.array(value)
            v_shp = value.shape
            if len(v_shp) == 1 and v_shp[0] > 0:
                raise AttributeError(msg.err2003)
            # If len(v_shp) == 1 and v_shp[0] == 0, then param_val was 'None'
            # and has been set to np.empty((0,)) by default
        self.__dict__[key] = value


class CatParam(CatMType):
    """ This class is used to define a categorized 1-dimensional parameter map
    that is associated to a model type (e.g. `model`, etc). The categories of
    each model parameter in `param_val` is indicated in `param_cat`. These
    categories are automatically updated if `param_val` or `categories` change.


    Notes
    -----
        The shape of `param_val` and `param_cat` is `shape=(nparam,)`.

    """

    _nint_cat = 16
    _dtype_cat = 'int' + str(_nint_cat)

    def __init__(self, dtype_val='float64', param_val=None, categories=None):
        # Assuring instance attributes 'param_val', 'categories' and
        # 'param_cat'

        # Check if _dtype_cat is sufficient
        if categories is not None \
                and len(categories) > 2**(self._nint_cat - 1):
            raise AttributeError(msg.err2004(len(categories)))

        super().__init__(dtype_val, param_val, categories)

    def __setattr__(self, key, value):
        """ The '__setattr__' method is customized for updating the private
        '__param_cat' attribute whenever 'param_val' is set.
        """
        update = False
        if key == 'param_val':
            # Avoid wrong definitions
            if not isinstance(value, np.ndarray):
                value = np.array(value)
            if len(value.shape) > 1:
                raise AttributeError(msg.err2002)
            # Check whether 'param_cat' can be updated
            if hasattr(self, 'categories') and len(self.categories) > 0:
                update = True
        elif key == 'categories':
            # Avoid wrong type
            if not isinstance(value, list):
                raise AttributeError(msg.err2005)
            # Check whether 'param_cat' can be updated
            if hasattr(self, 'param_val') and self.param_val is not None:
                update = True

        # Set value by '__setattr__' of 'Dict'
        self.__dict__[key] = value

        # Update 'param_cat' if needed
        if update:
            self._update_param_cat()

    def _update_param_cat(self):
        """ This function is used to update '__param_cat' which indicates the
        category index for each parameter value.

        Returns
        ------
        None

        """
        # Categorize the parameters according to self.categories
        self.__param_cat = np.zeros_like(self.param_val, dtype=self._dtype_cat)
        test_cat = np.zeros_like(self.param_val, dtype=bool)
        for icat, cat in enumerate(self.categories):
            ind = [np.logical_and(self.param_val >= intvl[0],
                                  self.param_val < intvl[1])
                   for intvl in cat]
            cat_ind = np.logical_or.reduce(ind)
            self.__param_cat += cat_ind * icat
            test_cat = np.logical_or(test_cat, cat_ind)

        # Check if all parameters have been categorized
        if not np.all(test_cat):
            ind = np.where(np.logical_not(test_cat))[0]
            raise ValueError(msg.err2001(ind,
                                         self.param_val[ind],
                                         self.categories))

    @property
    def param_cat(self):
        """ Category indicator array.

        Returns
        -------
        ndarray, shape=(nparam,)
            Category indicators of the values in `param_val`.
        """
        return self.__param_cat



""" 'cnsts.py' is dedicated to the definitions of the main PoPEx constants.
"""

# Tolerance of machine precision
# ------------------------------
NP_MIN_TOL = 1e-12

# Bounds for optimization
# -----------------------
ALPHA_MIN = 0.01
ALPHA_MAX = 100

# Job control
# -----------
NRERUN = 3

# Recompute p_cat frequency
# -------------------------
RECMP_P_CAT_FREQ = 1000


#!/usr/bin/env python

# Generic dependencies
from setuptools import setup, find_packages
import unittest


# Get unittest functions
def test_suite():
    test_loader = unittest.TestLoader()
    return test_loader.discover('tests', pattern='test_*.py')


# Long description from the README.md file
with open("README.md", "r") as file_handle:
    long_description = file_handle.read()


# Generic requirements
install_requires = ['numpy', 'scipy']

_setup_data = {
    'name':                 'popex',
    'version':              '1.0.0',
    'description':          'PoPEx sampling package',
    'long_description':      long_description,
    'long_description_content_type': 'text/markdown',

    'author':               'Christoph Jaeggli',
    'author_email':         'christoph.jaeggli@gmail.com',
    'url':                  'https://github.com/randlab/PoPEx',

    'license':              'MIT',

    'packages':             find_packages(),
    'install_requires':     install_requires,
}

if __name__ == '__main__':
    setup(**_setup_data)
